import torch
import torch.nn.functional as F
from torch.types import _dtype as DType
from typing import Any, Callable, Optional

__all__ = [
    'masked_softmax',
    'masked_sum',
    'masked_mean',
    'masked_max',
    'masked_min'
]

def _fill_with_mask(input: torch.Tensor, mask: torch.Tensor, fill_value) -> torch.Tensor:
    inverted_mask = (1.0 - mask.float()).bool()
    return input.masked_fill(inverted_mask, fill_value)

def _call_torch(func: Callable, **kwargs) -> Any:
    if kwargs["dim"] is None:
        kwargs.pop("dim")
        kwargs.pop("keepdim")
    return func(**kwargs)

def masked_softmax(
    input: torch.Tensor,
    mask: torch.Tensor,
    dim: Optional[int] = None,
    _stacklevel: int = 3,
    dtype: Optional[DType] = None
) -> torch.Tensor:
    if mask is None:
        return F.softmax(input, dim=dim, _stacklevel=_stacklevel, dtype=dtype)

    masked_input = _fill_with_mask(input, mask, -float('inf'))
    return F.softmax(masked_input, dim=dim, _stacklevel=_stacklevel, dtype=dtype)

def masked_sum(
    input: torch.Tensor,
    mask: torch.Tensor,
    dim: Optional[int] = None,
    keepdim: bool = False,
    dtype: Optional[DType] = None
) -> torch.Tensor:
    if mask is None:
        return _call_torch(input.sum, dim=dim, keepdim=keepdim, dtype=dtype)

    masked_input = _fill_with_mask(input, mask, 0.)
    return _call_torch(masked_input.sum, dim=dim, keepdim=keepdim, dtype=dtype)

def masked_mean(
    input: torch.Tensor,
    mask: torch.Tensor,
    dim: Optional[int] = None,
    keepdim: bool = False,
    dtype: Optional[DType] = None
) -> torch.Tensor:
    if mask is None:
        return _call_torch(input.mean, dim=dim, keepdim=keepdim, dtype=dtype)

    mask_sum = _call_torch(mask.float().sum, dim=dim, keepdim=keepdim)
    mask_sum = mask_sum.clamp(min=1.).to(dtype)

    return masked_sum(input, mask, dim, keepdim, dtype) / mask_sum

def masked_max(
    input: torch.Tensor,
    mask: torch.Tensor,
    dim: Optional[int] = None,
    keepdim: bool = False
) -> torch.Tensor:
    if mask is None:
        return _call_torch(input.max, dim=dim, keepdim=keepdim)

    masked_input = _fill_with_mask(input, mask, -float('inf'))
    return _call_torch(masked_input.max, dim=dim, keepdim=keepdim)

def masked_min(
    input: torch.Tensor,
    mask: torch.Tensor,
    dim: Optional[int] = None,
    keepdim: bool = False
) -> torch.Tensor:
    if mask is None:
        return _call_torch(input.min, dim=dim, keepdim=keepdim)

    masked_input = _fill_with_mask(input, mask, float('inf'))
    return _call_torch(masked_input.min, dim=dim, keepdim=keepdim)



from os import path
from setuptools import setup, find_packages

current_path = path.abspath(path.dirname(__file__))

def readme():
    readme_path = path.join(current_path, "README.md")
    with open(readme_path, "r", encoding="utf-8") as fp:
        return fp.read()

def dependencies():
    dep_path = path.join(current_path, "requirements.txt")
    with open(dep_path, "r", encoding="utf-8") as fp:
        return list(map(lambda x: x.strip(), fp.readlines()))

setup(
    name = "torchmasked",
    version = "0.1.0",
    packages = find_packages(),
    description = "Tensor operations with mask for PyTorch.",
    long_description = readme(),
    long_description_content_type = "text/markdown",
    classifiers = [
        "Intended Audience :: Science/Research",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    author = "Xiaohan Zou",
    license = 'MIT',
    author_email = "renovamenzxh@gmail.com",
    url = "https://github.com/Renovamen/torchmasked",
    python_requires = ">=3.6",
    install_requires = dependencies(),
    extras_require = {
        'dev': [
            'numpy>=1.14.0'
        ]
    }
)



import torch
from torch import nn
from typing import Optional

import torchmasked.functional as F

__all__ = ['MaskedSoftmax']

class MaskedSoftmax(nn.Module):
    def __init__(self, dim: Optional[int] = None) -> None:
        super(MaskedSoftmax, self).__init__()
        self.dim = dim

    def __setstate__(self, state):
        self.__dict__.update(state)
        if not hasattr(self, 'dim'):
            self.dim = None

    def forward(self, input: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        return F.masked_softmax(input, mask, dim=self.dim, _stacklevel=5)

    def extra_repr(self) -> str:
        return 'dim={dim}'.format(dim=self.dim)
